@@ -33,22 +33,24 @@
 enum {
 	PROP_0,
 	PROP_COMIC,
+	PROP_ZOOM_MODE,
 	PROP_SCALE
 };
 
 struct _BuohViewComicPrivate {
 	BuohView        *view;
 	GtkWidget       *image;
-	GdkPixbuf       *current;
+	GString         *comic_data;
 
 	BuohComic       *comic;
+	BuohViewZoomMode zoom_mode;
 	gdouble          scale;
 
-	guint            load_monitor;
 	BuohComicLoader *comic_loader;
+	GdkPixbufLoader *pixbuf_loader;
 };
 
-static GtkTargetEntry targets[] = {
+static const GtkTargetEntry targets[] = {
 	{ "text/uri-list", 0, 0 },
 	{ "x-url/http",    0, 0 }
 };
@@ -60,13 +62,14 @@ static GtkTargetEntry targets[] = {
 #define ZOOM_OUT_FACTOR (1.0 / ZOOM_IN_FACTOR)
 
 #define MIN_ZOOM_SCALE 0.6
-#define MAX_ZOOM_SCALE 2
+#define MAX_ZOOM_SCALE 4
 
-static GtkViewportClass *parent_class = NULL;
+#define DATA_SIZE 61440 /* 60K */
 
 static void     buoh_view_comic_init                  (BuohViewComic *m_view);
 static void     buoh_view_comic_class_init            (BuohViewComicClass *klass);
 static void     buoh_view_comic_finalize              (GObject          *object);
+static void     buoh_view_comic_dispose               (GObject          *object);
 static void     buoh_view_comic_set_property          (GObject          *object,
 						       guint             prop_id,
 						       const GValue     *value,
@@ -79,6 +82,10 @@ static gboolean buoh_view_comic_key_press_event       (GtkWidget        *widget,
 						       GdkEventKey      *event);
 static gboolean buoh_view_comic_scroll_event          (GtkWidget        *widget,
 						       GdkEventScroll   *event);
+static void     buoh_view_comic_size_allocate         (GtkWidget        *widget,
+						       GtkAllocation    *allocation);
+static void     buoh_view_comic_size_request          (GtkWidget        *widget,
+						       GtkRequisition   *requisition);
 static void     buoh_view_comic_drag_begin            (GtkWidget        *widget,
 						       GdkDragContext   *drag_context,
 						       gpointer          gdata);
@@ -95,38 +102,19 @@ static void     bouh_view_comic_view_status_changed   (GObject          *object,
 						       GParamSpec       *arg,
 						       gpointer          gdata);
 static void     buoh_view_comic_prepare_load          (BuohViewComic    *c_view);
-static void     buoh_view_comic_set_image_from_pixbuf (BuohViewComic    *c_view,
-						       GdkPixbuf        *pixbuf);
-static gboolean buoh_view_comic_load_monitor          (gpointer          gdata);
+static void     buoh_view_comic_load_finished         (BuohViewComic    *c_view,
+						       gpointer          gdata);
 static void     buoh_view_comic_load                  (BuohViewComic    *c_view);
+static void     buoh_view_comic_update_scrollbar_policy (BuohViewComic  *c_view);
+static gdouble  buoh_view_comic_get_scale_for_width   (BuohViewComic    *c_view,
+						       gint              width);
+static gdouble  buoh_view_comic_get_scale_for_height  (BuohViewComic    *c_view,
+						       gint              height);
 static void     buoh_view_comic_zoom                  (BuohViewComic    *c_view,
 						       gdouble           factor,
 						       gboolean          relative);
 
-GType
-buoh_view_comic_get_type (void)
-{
-	static GType type = 0;
-
-	if (!type) {
-		static const GTypeInfo info = {
-			sizeof (BuohViewComicClass),
-			(GBaseInitFunc) NULL,
-			(GBaseFinalizeFunc) NULL,
-			(GClassInitFunc) buoh_view_comic_class_init,
-			NULL,
-			NULL,
-			sizeof (BuohViewComic),
-			0,
-			(GInstanceInitFunc) buoh_view_comic_init
-		};
-
-		type = g_type_register_static (GTK_TYPE_VIEWPORT, "BuohViewComic",
-					       &info, 0);
-	}
-
-	return type;
-}
+G_DEFINE_TYPE (BuohViewComic, buoh_view_comic, GTK_TYPE_VIEWPORT)
 
 static void
 buoh_view_comic_init (BuohViewComic *c_view)
@@ -135,11 +123,10 @@ buoh_view_comic_init (BuohViewComic *c_view)
 	
 	c_view->priv = BUOH_VIEW_COMIC_GET_PRIVATE (c_view);
 
-	c_view->priv->view = NULL;
+	c_view->priv->zoom_mode = VIEW_ZOOM_FIT_WIDTH;
 	c_view->priv->scale = 1.0;
-	c_view->priv->comic = NULL;
-	c_view->priv->load_monitor = 0;
 	c_view->priv->comic_loader = buoh_comic_loader_new ();
+	c_view->priv->comic_data = g_string_sized_new (DATA_SIZE);
 	
 	c_view->priv->image = gtk_image_new ();
 	gtk_container_add (GTK_CONTAINER (c_view),
@@ -159,6 +146,11 @@ buoh_view_comic_init (BuohViewComic *c_view)
 			  "drag-data-get",
 			  G_CALLBACK (buoh_view_comic_drag_data_get),
 			  NULL);
+
+	g_signal_connect_swapped (G_OBJECT (c_view->priv->comic_loader),
+				  "finished",
+				  G_CALLBACK (buoh_view_comic_load_finished),
+				  (gpointer) c_view);
 	
 	gtk_widget_show (GTK_WIDGET (c_view));
 }
@@ -174,8 +166,8 @@ buoh_view_comic_class_init (BuohViewComicClass *klass)
 
 	widget_class->key_press_event = buoh_view_comic_key_press_event;
 	widget_class->scroll_event = buoh_view_comic_scroll_event;
-
-	parent_class = g_type_class_peek_parent (klass);
+	widget_class->size_allocate = buoh_view_comic_size_allocate;
+	widget_class->size_request = buoh_view_comic_size_request;
 
 	/* Properties */
 	g_object_class_install_property (object_class,
@@ -184,19 +176,28 @@ buoh_view_comic_class_init (BuohViewComicClass *klass)
 							       "Comic",
 							       "The current comic",
 							       G_PARAM_READWRITE));
+	g_object_class_install_property (object_class,
+					 PROP_ZOOM_MODE,
+					 g_param_spec_enum ("zoom-mode",
+							    "ZoomMode",
+							    "The view zoom mode",
+							    BUOH_TYPE_VIEW_ZOOM_MODE,
+							    VIEW_ZOOM_FIT_WIDTH,
+							    G_PARAM_READWRITE));
 	g_object_class_install_property (object_class,
 					 PROP_SCALE,
 					 g_param_spec_double ("scale",
 							      "Scale",
 							      "Current view scale",
-							      MIN_ZOOM_SCALE,
-							      MAX_ZOOM_SCALE,
+							      G_MINDOUBLE,
+							      G_MAXDOUBLE,
 							      1.0,
 							      G_PARAM_READWRITE));
 	
 	g_type_class_add_private (klass, sizeof (BuohViewComicPrivate));
 
 	object_class->finalize = buoh_view_comic_finalize;
+	object_class->dispose = buoh_view_comic_dispose;
 }
 
 static void
@@ -206,18 +207,36 @@ buoh_view_comic_finalize (GObject *object)
 
 	buoh_debug ("buoh-view-comic finalize");
 
-	if (c_view->priv->load_monitor > 0) {
-		g_source_remove (c_view->priv->load_monitor);
-		c_view->priv->load_monitor = 0;
+	if (c_view->priv->comic_data) {
+		g_string_free (c_view->priv->comic_data, TRUE);
+		c_view->priv->comic_data = NULL;
+	}
+	
+	if (c_view->priv->pixbuf_loader) {
+		gdk_pixbuf_loader_close (c_view->priv->pixbuf_loader, NULL);
+		g_object_unref (c_view->priv->pixbuf_loader);
+		c_view->priv->pixbuf_loader = NULL;
 	}
 
+	if (G_OBJECT_CLASS (buoh_view_comic_parent_class)->finalize)
+		(* G_OBJECT_CLASS (buoh_view_comic_parent_class)->finalize) (object);
+}
+
+static void
+buoh_view_comic_dispose (GObject *object)
+{
+	BuohViewComic *c_view = BUOH_VIEW_COMIC (object);
+
 	if (c_view->priv->comic_loader) {
+		g_signal_handlers_disconnect_by_func (c_view->priv->comic_loader,
+						      buoh_view_comic_load_finished,
+						      c_view);
 		g_object_unref (c_view->priv->comic_loader);
 		c_view->priv->comic_loader = NULL;
 	}
-	
-	if (G_OBJECT_CLASS (parent_class)->finalize)
-		(* G_OBJECT_CLASS (parent_class)->finalize) (object);
+
+	if (G_OBJECT_CLASS (buoh_view_comic_parent_class)->dispose)
+		(* G_OBJECT_CLASS (buoh_view_comic_parent_class)->dispose) (object);
 }
 
 static void
@@ -232,6 +251,11 @@ buoh_view_comic_set_property (GObject      *object,
 	case PROP_COMIC:
 		c_view->priv->comic = BUOH_COMIC (g_value_get_pointer (value));
 
+		break;
+	case PROP_ZOOM_MODE:
+		c_view->priv->zoom_mode = g_value_get_enum (value);
+		buoh_view_comic_update_scrollbar_policy (c_view);
+		
 		break;
 	case PROP_SCALE:
 		c_view->priv->scale = g_value_get_double (value);
@@ -254,6 +278,10 @@ buoh_view_comic_get_property (GObject    *object,
 	case PROP_COMIC:
 		g_value_set_pointer (value, c_view->priv->comic);
 
+		break;
+	case PROP_ZOOM_MODE:
+		g_value_set_enum (value, c_view->priv->zoom_mode);
+
 		break;
 	case PROP_SCALE:
 		g_value_set_double (value, c_view->priv->scale);
@@ -341,6 +369,74 @@ buoh_view_comic_scroll_event (GtkWidget *widget, GdkEventScroll *event)
 	return FALSE;
 }
 
+static gboolean
+buoh_view_comic_update_zoom_cb (BuohViewComic *c_view)
+{
+	GdkPixbuf *pixbuf;
+	gdouble    new_scale;
+
+	pixbuf = buoh_comic_get_pixbuf (c_view->priv->comic);
+	if (!pixbuf)
+		return FALSE;
+
+	switch (c_view->priv->zoom_mode) {
+	case VIEW_ZOOM_FREE:
+		new_scale = c_view->priv->scale;
+		break;
+	case VIEW_ZOOM_BEST_FIT: {
+		gdouble scale_width;
+		gdouble scale_height;
+
+		scale_width =
+			buoh_view_comic_get_scale_for_width (c_view,
+							     gdk_pixbuf_get_width (pixbuf));
+		
+		scale_height =
+			buoh_view_comic_get_scale_for_height (c_view,
+							      gdk_pixbuf_get_height (pixbuf));
+
+		new_scale = MIN (scale_width, scale_height);
+	}
+		break;
+	case VIEW_ZOOM_FIT_WIDTH:
+		new_scale =
+			buoh_view_comic_get_scale_for_width (c_view,
+							     gdk_pixbuf_get_width (pixbuf));
+		break;
+	default:
+		g_assert_not_reached ();
+
+	}
+
+	if (new_scale != c_view->priv->scale)
+		buoh_view_comic_zoom (c_view, new_scale, FALSE);
+	
+	return FALSE;
+}
+
+static void
+buoh_view_comic_size_request (GtkWidget *widget, GtkRequisition *requisition)
+{
+	requisition->width = -1;
+	requisition->height = -1;
+}
+
+static void
+buoh_view_comic_size_allocate (GtkWidget *widget, GtkAllocation *allocation)
+{
+	BuohViewComic *c_view = BUOH_VIEW_COMIC (widget);
+	static gint    id = 0;
+
+	if (c_view->priv->comic) {
+		if (id > 0)
+			g_source_remove (id);
+		id = g_idle_add ((GSourceFunc) buoh_view_comic_update_zoom_cb,
+				 c_view);
+	}
+	
+	GTK_WIDGET_CLASS (buoh_view_comic_parent_class)->size_allocate (widget, allocation);
+}
+
 GtkWidget *
 buoh_view_comic_new (BuohView *view)
 {
@@ -356,7 +452,7 @@ buoh_view_comic_new (BuohView *view)
 			  "notify::status",
 			  G_CALLBACK (bouh_view_comic_view_status_changed),
 			  (gpointer) c_view);
-	
+
 	return c_view;
 }
 
@@ -381,7 +477,7 @@ buoh_view_comic_drag_data_get (GtkWidget *widget, GdkDragContext *drag_context,
 			       gpointer gdata)
 {
 	BuohViewComic *c_view = BUOH_VIEW_COMIC (widget);
-	gchar         *uri = NULL;
+	const gchar   *uri;
 
 	uri = buoh_comic_get_uri (c_view->priv->comic);
 	if (uri) {
@@ -390,7 +486,64 @@ buoh_view_comic_drag_data_get (GtkWidget *widget, GdkDragContext *drag_context,
 					8,
 					(guchar *)uri,
 					strlen (uri));
-		g_free (uri);
+	}
+}
+
+static void
+buoh_view_comic_size_prepared (GdkPixbufLoader *loader,
+			       gint             width,
+			       gint             height,
+			       BuohViewComic   *c_view)
+{
+	switch (c_view->priv->zoom_mode) {
+	case VIEW_ZOOM_BEST_FIT: {
+		gdouble scale_width;
+		gdouble scale_height;
+
+		scale_width = buoh_view_comic_get_scale_for_width (c_view, width);
+		
+		scale_height = buoh_view_comic_get_scale_for_height (c_view, height);
+		
+		c_view->priv->scale = MIN (scale_width, scale_height);
+	}
+		break;
+	case VIEW_ZOOM_FIT_WIDTH: {
+		GtkWidget *swindow;
+
+		/* We have to predict if a vscrollbar will be needed
+		 * so that we'll have to take it into account
+		 */
+		swindow = gtk_widget_get_parent (GTK_WIDGET (c_view));
+		if (GTK_IS_SCROLLED_WINDOW (swindow)) {
+			GtkRequisition req;
+			gint           scrollbar_spacing;
+			gint           new_scale;
+			gint           widget_width;
+			gint           widget_height;
+
+			widget_width = GTK_WIDGET (c_view)->allocation.width;
+			widget_width -= 2 * GTK_WIDGET (c_view)->style->xthickness;
+
+			new_scale = (gdouble)widget_width / (gdouble)width;
+
+			widget_height = GTK_WIDGET (c_view)->allocation.height;
+
+			if ((height * new_scale) > widget_height) {
+				gtk_widget_size_request (GTK_SCROLLED_WINDOW (swindow)->vscrollbar, &req);
+				gtk_widget_style_get (swindow,
+						      "scrollbar_spacing", &scrollbar_spacing,
+						      NULL);
+				
+				GTK_WIDGET (c_view)->allocation.width -= (req.width + scrollbar_spacing);
+			}
+		}
+		
+		c_view->priv->scale =
+			buoh_view_comic_get_scale_for_width (c_view, width);
+	}
+		break;
+	default:
+		break;
 	}
 }
 
@@ -408,6 +561,11 @@ buoh_view_comic_prepare_load (BuohViewComic *c_view)
 	gtk_adjustment_set_value (hadjustment, 0.0);
 	gtk_adjustment_set_value (vadjustment, 0.0);
 
+	buoh_view_comic_update_scrollbar_policy (c_view);
+
+	if (GTK_WIDGET_REALIZED (GTK_WIDGET (c_view)))
+		gdk_window_set_cursor (GTK_WIDGET (c_view)->window, NULL);
+
 #if GTK_CHECK_VERSION(2,8,0)
 	gtk_image_clear (GTK_IMAGE (c_view->priv->image));
 #else 
@@ -415,13 +573,16 @@ buoh_view_comic_prepare_load (BuohViewComic *c_view)
 				  GTK_STOCK_MISSING_IMAGE,
 				  GTK_ICON_SIZE_DIALOG);
 #endif
-	
 }
+
 static void
 bouh_view_comic_changed_comic_cb (GObject *object, GParamSpec *arg, gpointer gdata)
 {
 	BuohViewComic *c_view = BUOH_VIEW_COMIC (object);
 
+	/* Cancel current load if needed */
+	buoh_comic_loader_cancel (c_view->priv->comic_loader);
+	
 	buoh_view_comic_prepare_load (c_view);
 
 	buoh_view_comic_load (c_view);
@@ -444,7 +605,8 @@ bouh_view_comic_view_status_changed (GObject *object, GParamSpec *arg, gpointer
 }
 
 static void
-buoh_view_comic_set_image_from_pixbuf (BuohViewComic *c_view, GdkPixbuf *pixbuf)
+buoh_view_comic_set_image_from_pixbuf (BuohViewComic *c_view,
+				       GdkPixbuf     *pixbuf)
 {
 	GdkPixbuf *new_pixbuf;
 
@@ -461,161 +623,219 @@ buoh_view_comic_set_image_from_pixbuf (BuohViewComic *c_view, GdkPixbuf *pixbuf)
 	} else {
 		gtk_image_set_from_pixbuf (GTK_IMAGE (c_view->priv->image), pixbuf);
 	}
-
-	gtk_widget_show (c_view->priv->image);
 }
 
-static gboolean
-buoh_view_comic_load_monitor (gpointer gdata)
+static void
+buoh_view_comic_load_finished (BuohViewComic *c_view,
+			       gpointer       gdata)
 {
-	BuohViewComic    *c_view = BUOH_VIEW_COMIC (gdata);
-	static GdkCursor *cursor = NULL;
-	static gint       height = 0;
+	GError    *error = NULL;
+	GdkPixbuf *pixbuf;
 
-	switch (c_view->priv->comic_loader->status) {
-	case LOADER_STATE_READY:
-		return TRUE;
-	case LOADER_STATE_RUNNING:
-		/* Watch cursor */
-		if (!cursor) {
-			cursor = gdk_cursor_new (GDK_WATCH);
-			if (!GTK_WIDGET_REALIZED (c_view))
-				gtk_widget_realize (GTK_WIDGET (c_view));
-			gdk_window_set_cursor (GTK_WIDGET (c_view)->window, cursor);
-		}
+	buoh_debug ("buoh-view-comic-load finished");
 
-		if (GDK_IS_PIXBUF (c_view->priv->comic_loader->pixbuf)) {
-			gint new_height = gdk_pixbuf_get_height (c_view->priv->comic_loader->pixbuf);
-
-			if (new_height != height) {
-				g_mutex_lock (c_view->priv->comic_loader->pixbuf_mutex);
-				gtk_image_set_from_pixbuf (GTK_IMAGE (c_view->priv->image),
-							   c_view->priv->comic_loader->pixbuf);
-				g_mutex_unlock (c_view->priv->comic_loader->pixbuf_mutex);
-				
-				gtk_widget_show (c_view->priv->image);
-				height = new_height;
-			}
-		}
-
-		return TRUE;
-	case LOADER_STATE_FINISHED:
-		if (GDK_IS_PIXBUF (c_view->priv->comic_loader->pixbuf)) {
-			g_mutex_lock (c_view->priv->comic_loader->pixbuf_mutex);
-			gtk_image_set_from_pixbuf (GTK_IMAGE (c_view->priv->image),
-						   c_view->priv->comic_loader->pixbuf);
-			if (c_view->priv->scale == 1.0) {
-				/* We have both the compressed and uncompressed image.
-				 * Setting the pixbuf avoids the cache uncompressing
-				 * the image again and having a new pixbuf instead of a
-				 * reference.
-				 */
-				buoh_comic_set_pixbuf (c_view->priv->comic,
-						       c_view->priv->comic_loader->pixbuf);
-			}
-			buoh_comic_set_image (c_view->priv->comic,
-					      c_view->priv->comic_loader->image);
-			g_object_unref (c_view->priv->comic_loader->pixbuf);
-			g_mutex_unlock (c_view->priv->comic_loader->pixbuf_mutex);
-
-			gtk_widget_show (c_view->priv->image);
-			
-			g_object_set (G_OBJECT (c_view->priv->view),
-				      "status", STATE_COMIC_LOADED,
-				      NULL);
-		}
-	case LOADER_STATE_STOPPING:
-		if (cursor) {
-			gdk_window_set_cursor (GTK_WIDGET (c_view)->window, NULL);
-			gdk_cursor_unref (cursor);
-			cursor = NULL;
-		}
-
-		buoh_debug ("Monitor exit (stopping/finished)");
-
-		height = 0;
-
-		return FALSE;
-	case LOADER_STATE_FAILED:
-		if (cursor) {
-			gdk_window_set_cursor (GTK_WIDGET (c_view)->window, NULL);
-			gdk_cursor_unref (cursor);
-			cursor = NULL;
-		}
+	gdk_pixbuf_loader_close (c_view->priv->pixbuf_loader, NULL);
 
+	gdk_window_set_cursor (GTK_WIDGET (c_view)->window, NULL);
+	
+	buoh_comic_loader_get_error (c_view->priv->comic_loader, &error);
+	if (error) {
 		buoh_view_set_message_title (BUOH_VIEW (c_view->priv->view),
 					     _("Error Loading Comic"));
 		
-		if (c_view->priv->comic_loader->error) {
-			buoh_view_set_message_text (BUOH_VIEW (c_view->priv->view),
-						    c_view->priv->comic_loader->error->message);
-		} else {
-			buoh_view_set_message_text (BUOH_VIEW (c_view->priv->view),
-						    _("There has been an error when loading the comic. "
-						      "It use to be due to an error on the remote server. "
-						      "Please, try again later."));
-		}
+		buoh_view_set_message_text (BUOH_VIEW (c_view->priv->view),
+					    error->message);
+		g_error_free (error);
 		
 		buoh_view_set_message_icon (BUOH_VIEW (c_view->priv->view),
 					    GTK_STOCK_DIALOG_ERROR);
-
+		
 		g_object_set (G_OBJECT (c_view->priv->view),
 			      "status", STATE_MESSAGE_ERROR,
 			      NULL);
 
-		buoh_debug ("Monitor exit (failed)");
-
-		height = 0;
+		g_object_unref (c_view->priv->pixbuf_loader);
+		c_view->priv->pixbuf_loader = NULL;
 		
-		return FALSE;
-	default:
-		buoh_debug ("Monitor exit (unknown)");
-
-		height = 0;
-		
-		return FALSE;
+		return;
 	}
+
+	pixbuf = gdk_pixbuf_loader_get_pixbuf (c_view->priv->pixbuf_loader);
+	if (pixbuf) {
+		buoh_view_comic_set_image_from_pixbuf (c_view, pixbuf);
+		
+		if (c_view->priv->scale == 1.0) {
+			/* We have both the compressed and uncompressed image.
+			 * By setting the pixbuf to cache we avoid uncompressing
+			 * the image again and having a new pixbuf instead of a
+			 * reference.
+			 */
+			buoh_comic_set_pixbuf (c_view->priv->comic, pixbuf);
+		}
+
+		g_object_set (G_OBJECT (c_view->priv->view),
+			      "status", STATE_COMIC_LOADED,
+			      NULL);
+	}
+
+	if (c_view->priv->comic_data->len > 0) {
+		BuohComicImage *comic_image;
+
+		comic_image = g_new0 (BuohComicImage, 1);
+
+		comic_image->size = c_view->priv->comic_data->len;
+		comic_image->data = (guchar *) g_memdup (c_view->priv->comic_data->str,
+							 c_view->priv->comic_data->len);
+		
+		buoh_comic_set_image (c_view->priv->comic, comic_image);
+
+		c_view->priv->comic_data->len = 0;
+	}
+	
+	g_object_unref (c_view->priv->pixbuf_loader);
+	c_view->priv->pixbuf_loader = NULL;
+}
+
+static gboolean
+buoh_view_comic_set_image_from_loader (BuohViewComic *c_view)
+{
+	GdkPixbuf *pixbuf;
+	
+	pixbuf = gdk_pixbuf_loader_get_pixbuf (c_view->priv->pixbuf_loader);
+	if (pixbuf)
+		buoh_view_comic_set_image_from_pixbuf (c_view, pixbuf);
+
+	return FALSE;
+}
+
+static void
+buoh_view_comic_load_cb (const gchar   *data,
+			 guint          len,
+			 BuohViewComic *c_view)
+{
+	static guint id = 0;
+	GError *error = NULL;
+
+	gdk_pixbuf_loader_write (c_view->priv->pixbuf_loader,
+				 (guchar *)data, len, &error);
+
+	if (error) {
+		g_warning (error->message);
+		g_error_free (error);
+		return;
+	}
+
+	c_view->priv->comic_data = g_string_append_len (c_view->priv->comic_data,
+							data, len);
+
+	if (id > 0)
+		g_source_remove (id);
+	id = g_idle_add ((GSourceFunc) buoh_view_comic_set_image_from_loader,
+			 (gpointer) c_view);
 }
 
 static void
 buoh_view_comic_load (BuohViewComic *c_view)
 {
 	GdkPixbuf *pixbuf;
-	gchar     *uri;
-
-	if (c_view->priv->comic_loader->status == LOADER_STATE_RUNNING) {
-		buoh_debug ("Load already running");
-		buoh_comic_loader_stop (c_view->priv->comic_loader);
-		buoh_debug ("waiting thread");
-		buoh_comic_loader_wait (c_view->priv->comic_loader);
-		buoh_debug ("died");
-	}
 
 	pixbuf = buoh_comic_get_pixbuf (c_view->priv->comic);
 	if (pixbuf) {
+		buoh_view_comic_update_zoom_cb (c_view);
 		buoh_view_comic_set_image_from_pixbuf (c_view, pixbuf);
 		g_object_set (G_OBJECT (c_view->priv->view),
 			      "status", STATE_COMIC_LOADED,
 			      NULL);
 	} else {
+		GdkCursor *cursor;
+		
 		g_object_set (G_OBJECT (c_view->priv->view),
 			      "status", STATE_COMIC_LOADING,
 			      NULL);
-		uri = buoh_comic_get_uri (c_view->priv->comic);
-		buoh_comic_loader_run (c_view->priv->comic_loader,
-				       uri,
-				       c_view->priv->scale);
-		g_free (uri);
 
-		if (c_view->priv->load_monitor > 0)
-			g_source_remove (c_view->priv->load_monitor);
-		c_view->priv->load_monitor =
-			g_timeout_add (60,
-				       (GSourceFunc)buoh_view_comic_load_monitor,
-				       (gpointer)c_view);
+		cursor = gdk_cursor_new (GDK_WATCH);
+		gdk_window_set_cursor (GTK_WIDGET (c_view)->window, cursor);
+		gdk_cursor_unref (cursor);
+
+		if (c_view->priv->pixbuf_loader) {
+			gdk_pixbuf_loader_close (c_view->priv->pixbuf_loader, NULL);
+			g_object_unref (c_view->priv->pixbuf_loader);
+		}
+		c_view->priv->pixbuf_loader = gdk_pixbuf_loader_new ();
+		g_signal_connect (G_OBJECT (c_view->priv->pixbuf_loader),
+				  "size-prepared",
+				  G_CALLBACK (buoh_view_comic_size_prepared),
+				  (gpointer) c_view);
+
+		c_view->priv->comic_data->len = 0;
+		
+		buoh_comic_loader_load_comic (c_view->priv->comic_loader,
+					      c_view->priv->comic,
+					      (BuohComicLoaderLoadFunc) buoh_view_comic_load_cb,
+					      (gpointer) c_view);
 	}
 }
 
+static void
+buoh_view_comic_update_scrollbar_policy (BuohViewComic *c_view)
+{
+	GtkWidget *swindow;
+	
+	swindow = gtk_widget_get_parent (GTK_WIDGET (c_view));
+	if (!GTK_IS_SCROLLED_WINDOW (swindow))
+		return;
+
+	switch (c_view->priv->zoom_mode) {
+	case VIEW_ZOOM_FREE:
+		gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (swindow),
+						GTK_POLICY_AUTOMATIC,
+						GTK_POLICY_AUTOMATIC);
+		break;
+	case VIEW_ZOOM_FIT_WIDTH:
+		gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (swindow),
+						GTK_POLICY_NEVER,
+						GTK_POLICY_AUTOMATIC);
+		break;
+	case VIEW_ZOOM_BEST_FIT:
+		gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (swindow),
+						GTK_POLICY_NEVER,
+						GTK_POLICY_NEVER);
+		break;
+	}
+}
+
+static gdouble
+buoh_view_comic_get_scale_for_width (BuohViewComic *c_view,
+				     gint           width)
+{
+	GtkWidget *widget = GTK_WIDGET (c_view);
+	gint       widget_width;
+	gdouble    new_scale;
+
+	widget_width = widget->allocation.width;
+	widget_width -= 2 * widget->style->xthickness;
+	
+	new_scale = (gdouble)widget_width / (gdouble)width;
+
+	return new_scale;
+}
+
+static gdouble
+buoh_view_comic_get_scale_for_height (BuohViewComic *c_view,
+				      gint           height)
+{
+	GtkWidget *widget = GTK_WIDGET (c_view);
+	gint       widget_height;
+	gdouble    new_scale;
+
+	widget_height = widget->allocation.height;
+	widget_height -= 2 * widget->style->ythickness;
+	
+	new_scale = (gdouble)widget_height / (gdouble)height;
+
+	return new_scale;
+}
+
 static void
 buoh_view_comic_zoom (BuohViewComic *c_view, gdouble factor, gboolean relative)
 {
@@ -630,10 +850,13 @@ buoh_view_comic_zoom (BuohViewComic *c_view, gdouble factor, gboolean relative)
 	else
 		scale = factor;
 
+	if (scale == c_view->priv->scale)
+		return;
+
 	g_object_set (G_OBJECT (c_view),
-		      "scale", CLAMP (scale, MIN_ZOOM_SCALE, MAX_ZOOM_SCALE),
+		      "scale", scale,
 		      NULL);
-	
+
 	pixbuf = buoh_comic_get_pixbuf (c_view->priv->comic);
 	if (pixbuf) {
 		buoh_view_comic_set_image_from_pixbuf (c_view, pixbuf);
@@ -645,7 +868,7 @@ buoh_view_comic_is_min_zoom (BuohViewComic *c_view)
 {
 	g_return_val_if_fail (BUOH_IS_VIEW_COMIC (c_view), FALSE);
 	
-	return c_view->priv->scale == MIN_ZOOM_SCALE;
+	return c_view->priv->scale <= MIN_ZOOM_SCALE;
 }
 
 gboolean
@@ -653,14 +876,14 @@ buoh_view_comic_is_max_zoom (BuohViewComic *c_view)
 {
 	g_return_val_if_fail (BUOH_IS_VIEW_COMIC (c_view), FALSE);
 	
-	return c_view->priv->scale == MAX_ZOOM_SCALE;
+	return c_view->priv->scale >= MAX_ZOOM_SCALE;
 }
 
 gboolean
 buoh_view_comic_is_normal_size (BuohViewComic *c_view)
 {
 	g_return_val_if_fail (BUOH_IS_VIEW_COMIC (c_view), FALSE);
-	
+
 	return c_view->priv->scale == 1.0;
 }
 
@@ -668,7 +891,8 @@ void
 buoh_view_comic_zoom_in (BuohViewComic *c_view)
 {
 	g_return_if_fail (BUOH_IS_VIEW_COMIC (c_view));
-	
+
+	buoh_view_comic_set_zoom_mode (c_view, VIEW_ZOOM_FREE);
 	buoh_view_comic_zoom (c_view, ZOOM_IN_FACTOR, TRUE);
 }
 
@@ -676,7 +900,8 @@ void
 buoh_view_comic_zoom_out (BuohViewComic *c_view)
 {
 	g_return_if_fail (BUOH_IS_VIEW_COMIC (c_view));
-	
+
+	buoh_view_comic_set_zoom_mode (c_view, VIEW_ZOOM_FREE);
 	buoh_view_comic_zoom (c_view, ZOOM_OUT_FACTOR, TRUE);
 }
 
@@ -684,6 +909,44 @@ void
 buoh_view_comic_normal_size (BuohViewComic *c_view)
 {
 	g_return_if_fail (BUOH_IS_VIEW_COMIC (c_view));
-	
+
+	buoh_view_comic_set_zoom_mode (c_view, VIEW_ZOOM_FREE);
 	buoh_view_comic_zoom (c_view, 1.0, FALSE);
 }
+
+void
+buoh_view_comic_best_fit (BuohViewComic *c_view)
+{
+	g_return_if_fail (BUOH_IS_VIEW_COMIC (c_view));
+
+	buoh_view_comic_set_zoom_mode (c_view, VIEW_ZOOM_BEST_FIT);
+	gtk_widget_queue_resize (GTK_WIDGET (c_view));
+}
+
+void
+buoh_view_comic_fit_width (BuohViewComic *c_view)
+{
+	g_return_if_fail (BUOH_IS_VIEW_COMIC (c_view));
+
+	buoh_view_comic_set_zoom_mode (c_view, VIEW_ZOOM_FIT_WIDTH);
+	gtk_widget_queue_resize (GTK_WIDGET (c_view));
+}
+
+BuohViewZoomMode
+buoh_view_comic_get_zoom_mode (BuohViewComic *c_view)
+{
+	g_return_val_if_fail (BUOH_IS_VIEW_COMIC (c_view), 0);
+
+	return c_view->priv->zoom_mode;
+}
+
+void
+buoh_view_comic_set_zoom_mode (BuohViewComic   *c_view,
+			       BuohViewZoomMode mode)
+{
+	g_return_if_fail (BUOH_IS_VIEW_COMIC (c_view));
+	
+	g_object_set (G_OBJECT (c_view),
+		      "zoom-mode", mode,
+		      NULL);
+}
