@@ -58,10 +58,9 @@ struct _BuohWindowPrivate {
 
 #define GCONF_SHOW_TOOLBAR	"/apps/buoh/show_toolbar"
 #define GCONF_SHOW_STATUSBAR    "/apps/buoh/show_statusbar"
+#define GCONF_ZOOM_MODE         "/apps/buoh/zoom_mode"
 #define GCONF_LOCKDOWN_SAVE 	"/desktop/gnome/lockdown/disable_save_to_disk"
 
-static GtkWindowClass *parent_class = NULL;
-
 static void buoh_window_init                             (BuohWindow      *buoh_window);
 static void buoh_window_class_init                       (BuohWindowClass *klass);
 static void buoh_window_finalize                         (GObject         *object);
@@ -133,6 +132,10 @@ static void buoh_window_cmd_view_zoom_out               (GtkAction        *actio
 							 gpointer          gdata);
 static void buoh_window_cmd_view_zoom_normal            (GtkAction        *action,
 							 gpointer          gdata);
+static void buoh_window_cmd_view_zoom_best_fit          (GtkAction        *action,
+							 gpointer          gdata);
+static void buoh_window_cmd_view_zoom_fit_width         (GtkAction        *action,
+							 gpointer          gdata);
 static void buoh_window_cmd_go_previous                 (GtkAction        *action,
 							 gpointer          gdata);
 static void buoh_window_cmd_go_next                     (GtkAction        *action,
@@ -147,6 +150,7 @@ static void buoh_window_cmd_help_about                  (GtkAction        *actio
 							 gpointer          gdata);
 
 static void buoh_window_update_title                    (BuohWindow       *window);
+static void buoh_window_update_zoom_mode                (BuohWindow       *window);
 
 static const GtkActionEntry menu_entries[] = {
 
@@ -218,38 +222,20 @@ static const GtkToggleActionEntry menu_toggle_entries[] = {
 	  G_CALLBACK (buoh_window_cmd_view_toolbar), TRUE },
 	{ "ViewStatusbar", NULL, N_("St_atusbar"), NULL,
 	  N_("Changes the visibility of the statusbar"),
-	  G_CALLBACK (buoh_window_cmd_view_statusbar), TRUE }
+	  G_CALLBACK (buoh_window_cmd_view_statusbar), TRUE },
+	{ "ViewZoomBestFit", GTK_STOCK_ZOOM_FIT, N_("_Best Fit"), NULL,
+	  N_("Make the current comic fill the window"),
+	  G_CALLBACK (buoh_window_cmd_view_zoom_best_fit) },
+	{ "ViewZoomFitWidth", GTK_STOCK_ZOOM_FIT, N_("Fit Comic _Width"), NULL,
+	  N_("Make the current comic fill the window width"),
+	  G_CALLBACK (buoh_window_cmd_view_zoom_fit_width) }
 };
 
-GType
-buoh_window_get_type (void)
-{
-	static GType type = 0;
-
-	if (!type) {
-		static const GTypeInfo info = {
-			sizeof (BuohWindowClass),
-			(GBaseInitFunc) NULL,
-			(GBaseFinalizeFunc) NULL,
-			(GClassInitFunc) buoh_window_class_init,
-			NULL,
-			NULL,
-			sizeof (BuohWindow),
-			0,
-			(GInstanceInitFunc) buoh_window_init
-		};
-
-		type = g_type_register_static (GTK_TYPE_WINDOW, "BuohWindow",
-					       &info, 0);
-	}
-
-	return type;
-}
+G_DEFINE_TYPE (BuohWindow, buoh_window, GTK_TYPE_WINDOW)
 
 static void
 buoh_window_init (BuohWindow *buoh_window)
 {
-	gchar            *icon_path;
 	GtkWidget        *tree_view;
 	GtkTreeSelection *selection;
 	GtkWidget        *vbox, *paned, *menubar;
@@ -260,14 +246,12 @@ buoh_window_init (BuohWindow *buoh_window)
 	GError           *error = NULL;
 	gboolean          visible_toolbar;
 	gboolean          visible_statusbar;
+	BuohViewZoomMode  zoom_mode;
       
 	buoh_window->priv = BUOH_WINDOW_GET_PRIVATE (buoh_window);
 
 	gtk_window_set_title (GTK_WINDOW (buoh_window), "Buoh");
-
-	icon_path = g_build_filename (PIXMAPS_DIR, "buoh16x16.png", NULL);
-	gtk_window_set_icon_from_file (GTK_WINDOW (buoh_window), icon_path, NULL);
-	g_free (icon_path);
+	gtk_window_set_icon_name (GTK_WINDOW (buoh_window), "buoh");
 
 	buoh_window->priv->properties = NULL;
 	buoh_window->priv->add_dialog = NULL;
@@ -329,7 +313,7 @@ buoh_window_init (BuohWindow *buoh_window)
 	action = gtk_action_group_get_action (action_group, "ViewStatusbar");
 	gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action),
 				      visible_statusbar);
-	
+
 	/* Toolbar */
 	toolbar = gtk_ui_manager_get_widget (buoh_window->priv->ui_manager,
 					     "/Toolbar");
@@ -342,10 +326,15 @@ buoh_window_init (BuohWindow *buoh_window)
 	
 	/* Pane */
 	paned = gtk_hpaned_new ();
+	/* FIXME: Remember side position */
 	gtk_paned_set_position (GTK_PANED (paned), 230);
 
 	/* buoh view */
 	buoh_window->priv->view = BUOH_VIEW (buoh_view_new ());
+	zoom_mode = gconf_client_get_int (buoh_window->priv->gconf_client,
+					  GCONF_ZOOM_MODE,
+					  NULL);
+	buoh_view_set_zoom_mode (buoh_window->priv->view, zoom_mode);
 	g_signal_connect (G_OBJECT (buoh_window->priv->view), "notify::status",
 			  G_CALLBACK (buoh_window_view_status_change_cb),
 			  (gpointer) buoh_window);
@@ -405,6 +394,7 @@ buoh_window_init (BuohWindow *buoh_window)
 	buoh_window_comic_actions_set_sensitive (buoh_window, FALSE);
 	buoh_window_comic_save_to_disk_set_sensitive (buoh_window, FALSE);
 	buoh_window_set_sensitive (buoh_window, "ComicRemove", FALSE);
+	buoh_window_update_zoom_mode (buoh_window);
 	
 	gtk_widget_grab_focus (GTK_WIDGET (buoh_window->priv->view));
 
@@ -416,8 +406,6 @@ buoh_window_class_init (BuohWindowClass *klass)
 {
 	GObjectClass *object_class = G_OBJECT_CLASS (klass);
 
-	parent_class = g_type_class_peek_parent (klass);
-
 	g_type_class_add_private (klass, sizeof (BuohWindowPrivate));
 
 	object_class->finalize = buoh_window_finalize;
@@ -457,8 +445,8 @@ buoh_window_finalize (GObject *object)
 		buoh_window->priv->add_dialog = NULL;
 	}
 
-	if (G_OBJECT_CLASS (parent_class)->finalize)
-		(* G_OBJECT_CLASS (parent_class)->finalize) (object);
+	if (G_OBJECT_CLASS (buoh_window_parent_class)->finalize)
+		(* G_OBJECT_CLASS (buoh_window_parent_class)->finalize) (object);
 
 	buoh_exit_app (BUOH);
 }
@@ -495,42 +483,40 @@ buoh_window_cmd_comic_add (GtkAction *action, gpointer gdata)
 static void
 buoh_window_cmd_comic_remove (GtkAction *action, gpointer gdata)
 {
-	BuohWindow             *window = BUOH_WINDOW (gdata);
-	GtkTreeModel           *model = buoh_get_comics_model (BUOH);
-	GtkTreeIter             iter;
-	BuohComicManager       *cm;
-	BuohComicManager       *current_cm;
-	gchar                  *current_cm_id, *cm_id;
-	gboolean                valid;
+	BuohWindow       *window = BUOH_WINDOW (gdata);
+	GtkTreeModel     *model = buoh_get_comics_model (BUOH);
+	GtkTreeIter       iter;
+	BuohComicManager *cm;
+	BuohComicManager *current_cm;
+	const gchar      *current_cm_id;
+	const gchar      *cm_id;
+	gboolean          valid;
 
 	current_cm = buoh_comic_list_get_selected (window->priv->comic_list);
 
-	if (current_cm) {
-		current_cm_id = buoh_comic_manager_get_id (current_cm);
+	if (!current_cm)
+		return;
 
-		valid = gtk_tree_model_get_iter_first (model, &iter);
-
-		while (valid) {
-			gtk_tree_model_get (model, &iter,
-					    COMIC_LIST_COMIC_MANAGER, &cm,
+	current_cm_id = buoh_comic_manager_get_id (current_cm);
+	
+	valid = gtk_tree_model_get_iter_first (model, &iter);
+	
+	while (valid) {
+		gtk_tree_model_get (model, &iter,
+				    COMIC_LIST_COMIC_MANAGER, &cm,
+				    -1);
+		cm_id = buoh_comic_manager_get_id (cm);
+		g_object_unref (cm);
+		
+		if (g_ascii_strcasecmp (current_cm_id, cm_id) == 0) {
+			buoh_comic_list_clear_selection (window->priv->comic_list);
+			gtk_list_store_set (GTK_LIST_STORE (model), &iter,
+					    COMIC_LIST_VISIBLE, FALSE,
 					    -1);
-			cm_id = buoh_comic_manager_get_id (cm);
-			g_object_unref (cm);
-
-			if (g_ascii_strcasecmp (current_cm_id, cm_id) == 0) {
-				buoh_comic_list_clear_selection (window->priv->comic_list);
-				gtk_list_store_set (GTK_LIST_STORE (model), &iter,
-						    COMIC_LIST_VISIBLE, FALSE,
-						    -1);
-				valid = FALSE;
-			} else {
-				valid = gtk_tree_model_iter_next (model, &iter);
-			}
-
-			g_free (cm_id);
+			valid = FALSE;
+		} else {
+			valid = gtk_tree_model_iter_next (model, &iter);
 		}
-
-		g_free (current_cm_id);
 	}
 }
 
@@ -620,10 +606,10 @@ buoh_window_cmd_comic_copy_location (GtkAction *action, gpointer gdata)
 {
 	BuohWindow *window = BUOH_WINDOW (gdata);
 	BuohComic  *comic = buoh_view_get_comic (window->priv->view);
-	gchar      *uri;
 
 	if (comic) {
-		uri = buoh_comic_get_uri (comic);
+		const gchar *uri = buoh_comic_get_uri (comic);
+
 		buoh_debug ("Copy %s to clipboard", uri);
 
 		gtk_clipboard_set_text (gtk_clipboard_get (GDK_NONE), uri,
@@ -631,8 +617,6 @@ buoh_window_cmd_comic_copy_location (GtkAction *action, gpointer gdata)
 
 		gtk_clipboard_set_text (gtk_clipboard_get (GDK_SELECTION_PRIMARY), uri,
 					g_utf8_strlen (uri, -1));
-
-		g_free (uri);
 	}
 }
 
@@ -653,42 +637,35 @@ buoh_window_cmd_comic_properties (GtkAction *action, gpointer gdata)
 	BuohComicManager *cm  = buoh_comic_list_get_selected (window->priv->comic_list);
 	BuohComicManager *cm2 = NULL;
 	GtkWidget        *dialog;
-	gchar            *id1, *id2;
+	const gchar      *id1, *id2;
 	GList            *l = NULL;
 
-	if (cm) {
-		id1 = buoh_comic_manager_get_id (cm);
+	if (!cm)
+		return;
+	
+	id1 = buoh_comic_manager_get_id (cm);
+	
+	for (l = window->priv->properties; l; l = g_list_next (l)) {
+		cm2 = buoh_properties_dialog_get_comic_manager (
+			BUOH_PROPERTIES_DIALOG (l->data));
+		id2 = buoh_comic_manager_get_id (cm2);
 		
-		for (l = window->priv->properties; l; l = g_list_next (l)) {
-			cm2 = buoh_properties_dialog_get_comic_manager (
-				BUOH_PROPERTIES_DIALOG (l->data));
-			id2 = buoh_comic_manager_get_id (cm2);
-			
-			if (g_ascii_strcasecmp (id1, id2) == 0) {
-				gtk_window_present (GTK_WINDOW (l->data));
-				
-				g_free (id1);
-				g_free (id2);
-				
-				return;
-			}
-			
-			g_free (id2);
+		if (g_ascii_strcasecmp (id1, id2) == 0) {
+			gtk_window_present (GTK_WINDOW (l->data));
+			return;
 		}
-
-		g_free (id1);
-
-		dialog = buoh_properties_dialog_new ();
-		buoh_properties_dialog_set_comic_manager (BUOH_PROPERTIES_DIALOG (dialog), cm);
-		g_signal_connect (G_OBJECT (dialog), "destroy",
-				  G_CALLBACK (buoh_window_properties_dialog_destroyed),
-				  (gpointer) window);
-		gtk_window_set_transient_for (GTK_WINDOW (dialog), GTK_WINDOW (window));
-		
-		gtk_widget_show (dialog);
-
-		window->priv->properties = g_list_append (window->priv->properties, dialog);
 	}
+	
+	dialog = buoh_properties_dialog_new ();
+	buoh_properties_dialog_set_comic_manager (BUOH_PROPERTIES_DIALOG (dialog), cm);
+	g_signal_connect (G_OBJECT (dialog), "destroy",
+			  G_CALLBACK (buoh_window_properties_dialog_destroyed),
+			  (gpointer) window);
+	gtk_window_set_transient_for (GTK_WINDOW (dialog), GTK_WINDOW (window));
+	
+	gtk_widget_show (dialog);
+	
+	window->priv->properties = g_list_append (window->priv->properties, dialog);
 }
 
 static void
@@ -755,6 +732,42 @@ buoh_window_cmd_view_zoom_normal (GtkAction *action, gpointer gdata)
 	buoh_view_zoom_normal_size (window->priv->view);
 }
 
+static void
+buoh_window_cmd_view_zoom_best_fit (GtkAction *action, gpointer gdata)
+{
+	BuohWindow *window = BUOH_WINDOW (gdata);
+
+	if (gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action))) {
+		buoh_view_zoom_best_fit (window->priv->view);
+		gconf_client_set_int (window->priv->gconf_client,
+				      GCONF_ZOOM_MODE,
+				      VIEW_ZOOM_BEST_FIT, NULL);
+	} else {
+		buoh_view_zoom_normal_size (window->priv->view);
+		gconf_client_set_int (window->priv->gconf_client,
+				      GCONF_ZOOM_MODE,
+				      VIEW_ZOOM_FREE, NULL);
+	}
+}
+
+static void
+buoh_window_cmd_view_zoom_fit_width (GtkAction *action, gpointer gdata)
+{
+	BuohWindow *window = BUOH_WINDOW (gdata);
+
+	if (gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action))) {
+		buoh_view_zoom_fit_width (window->priv->view);
+		gconf_client_set_int (window->priv->gconf_client,
+				      GCONF_ZOOM_MODE,
+				      VIEW_ZOOM_FIT_WIDTH, NULL);
+	} else {
+		buoh_view_zoom_normal_size (window->priv->view);
+		gconf_client_set_int (window->priv->gconf_client,
+				      GCONF_ZOOM_MODE,
+				      VIEW_ZOOM_FREE, NULL);
+	}
+}
+
 static void
 buoh_window_cmd_go_previous (GtkAction *action, gpointer gdata)
 {
@@ -820,45 +833,34 @@ static void
 buoh_window_cmd_help_about (GtkAction *action, gpointer gdata)
 {
 	BuohWindow         *window = BUOH_WINDOW (gdata);
-	GdkPixbuf          *pixbuf;
-	gchar              *pixbuf_path;
 	static const gchar *authors[] = {
 		"Esteban Sánchez Muñoz <esteban@steve-o.org>",
 		"Pablo Arroyo Loma <zioma@linups.org>",
 		"Carlos Garcia Campos <carlosgc@gnome.org>",
-		                NULL
+		NULL
 	};
 
-	pixbuf_path = g_build_filename (PIXMAPS_DIR, "buoh64x64.png", NULL);
-	pixbuf = gdk_pixbuf_new_from_file_at_size (pixbuf_path, 48, 48, NULL);
-	g_free (pixbuf_path);
-
 	gtk_show_about_dialog (GTK_WINDOW (window),
 			       "name", _("Buoh online comics reader"),
 			       "version", VERSION,
 			       "copyright", "Copyright \xC2\xA9 2004 Esteban Sánchez Muñoz - Pablo Arroyo Loma",
 			       "authors", authors,
 			       "translator-credits", _("translator-credits"),
-			       "logo", pixbuf,
+			       "logo-icon-name", "buoh",
 			       NULL);
-
-	if (pixbuf)
-		g_object_unref (pixbuf);
 }
 
 static void
 buoh_window_update_title (BuohWindow *window)
 {
 	BuohComicManager *cm;
-	gchar            *cm_title;
-	gchar            *title;
+	gchar            *title = NULL;
 
 	cm = buoh_comic_list_get_selected (window->priv->comic_list);
 
 	if (cm) {
-		cm_title = buoh_comic_manager_get_title (cm);
-		title = g_strdup_printf ("%s - Buoh", cm_title);
-		g_free (cm_title);
+		title = g_strdup_printf ("%s - Buoh",
+					 buoh_comic_manager_get_title (cm));
 	} else {
 		title = g_strdup_printf ("Buoh");
 	}
@@ -906,7 +908,7 @@ buoh_window_comic_actions_set_sensitive (BuohWindow *window, gboolean sensitive)
 	buoh_window_comic_browsing_actions_set_sensitive (window, sensitive);
 
 	buoh_window_set_sensitive (window, "ComicProperties", sensitive);
-	buoh_window_set_sensitive (window, "ComicCopyURI",    sensitive); 
+	buoh_window_set_sensitive (window, "ComicCopyURI",    sensitive);
 	buoh_window_set_sensitive (window, "ViewZoomIn",
 				   buoh_view_is_max_zoom (window->priv->view) ?
 				   FALSE : sensitive);
@@ -916,6 +918,8 @@ buoh_window_comic_actions_set_sensitive (BuohWindow *window, gboolean sensitive)
 	buoh_window_set_sensitive (window, "ViewZoomNormal",
 				   buoh_view_is_normal_size (window->priv->view) ?
 				   FALSE : sensitive);
+	buoh_window_set_sensitive (window, "ViewZoomBestFit", sensitive);
+	buoh_window_set_sensitive (window, "ViewZoomFitWidth", sensitive);
 }
 
 static void
@@ -1002,9 +1006,8 @@ buoh_window_view_status_change_cb (GObject *object, GParamSpec *arg, gpointer gd
 	BuohComic        *comic = NULL;
 	BuohComicManager *cm;
 	GtkStatusbar     *statusbar = GTK_STATUSBAR (window->priv->statusbar);
-	gchar            *id, *title;
 	gchar            *message = NULL;
-
+	
 	cm = buoh_comic_list_get_selected (window->priv->comic_list);
 
 	switch (buoh_view_get_status (view)) {
@@ -1015,7 +1018,8 @@ buoh_window_view_status_change_cb (GObject *object, GParamSpec *arg, gpointer gd
 		break;
 	case STATE_MESSAGE_ERROR:
 		buoh_window_comic_actions_set_sensitive (window, FALSE);
-		buoh_window_comic_browsing_actions_set_sensitive (window, TRUE);
+		buoh_window_comic_browsing_actions_set_sensitive (window,
+								  TRUE);
 		break;
 	case STATE_COMIC_LOADING:
 		comic = buoh_view_get_comic (view);
@@ -1029,11 +1033,9 @@ buoh_window_view_status_change_cb (GObject *object, GParamSpec *arg, gpointer gd
 	case STATE_COMIC_LOADED:
 		comic = buoh_view_get_comic (view);
 
-		id = buoh_comic_get_id (comic);
-		title = buoh_comic_manager_get_title (cm);
-		message = g_strdup_printf ("%s - %s", title, id);
-		g_free (id);
-		g_free (title);
+		message = g_strdup_printf ("%s - %s",
+					   buoh_comic_manager_get_title (cm),
+					   buoh_comic_get_id (comic));
 		
 		buoh_window_comic_actions_set_sensitive (window,
 							 (comic) ? TRUE : FALSE);
@@ -1054,6 +1056,33 @@ buoh_window_view_status_change_cb (GObject *object, GParamSpec *arg, gpointer gd
 	gtk_widget_grab_focus (GTK_WIDGET (view));
 }
 
+static void
+buoh_window_update_zoom_mode (BuohWindow *window)
+{
+	BuohViewZoomMode  zoom_mode;
+	GtkAction        *action;
+
+	zoom_mode = buoh_view_get_zoom_mode (window->priv->view);
+	
+	action = gtk_action_group_get_action (window->priv->action_group,
+					      "ViewZoomBestFit");
+	g_signal_handlers_block_by_func
+		(action, G_CALLBACK (buoh_window_cmd_view_zoom_best_fit), window);
+	gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action),
+				      zoom_mode == VIEW_ZOOM_BEST_FIT);
+	g_signal_handlers_unblock_by_func
+		(action, G_CALLBACK (buoh_window_cmd_view_zoom_best_fit), window);
+	
+	action = gtk_action_group_get_action (window->priv->action_group,
+					      "ViewZoomFitWidth");
+	g_signal_handlers_block_by_func
+		(action, G_CALLBACK (buoh_window_cmd_view_zoom_fit_width), window);
+	gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (action),
+				      zoom_mode == VIEW_ZOOM_FIT_WIDTH);
+	g_signal_handlers_unblock_by_func
+		(action, G_CALLBACK (buoh_window_cmd_view_zoom_fit_width), window);
+}
+
 static void
 buoh_window_view_zoom_change_cb (BuohView *view, gpointer gdata)
 {
@@ -1066,6 +1095,8 @@ buoh_window_view_zoom_change_cb (BuohView *view, gpointer gdata)
 	buoh_window_set_sensitive (window, "ViewZoomNormal",
 				   !buoh_view_is_normal_size (window->priv->view));
 
+	buoh_window_update_zoom_mode (window);
+
 	gtk_widget_grab_focus (GTK_WIDGET (view));
 }
 
